@@ -60,24 +60,25 @@ func TestHttpsProxy(t *testing.T) {
 	}
 }
 
-func TestHttpsProxyTrustedProxies(t *testing.T) {
-	// Prepare the proxy server.
-	tlsConf, caPem := createServerTLSConfig(t)
-	dnsProxy := createTestProxy(t, tlsConf)
-
-	var gotAddr net.Addr
-	dnsProxy.RequestHandler = func(_ *Proxy, d *DNSContext) (err error) {
-		gotAddr = d.Addr
-
-		return dnsProxy.Resolve(d)
-	}
-
-	client := createTestHTTPClient(dnsProxy, caPem, false)
-
+func TestProxy_trustedProxies(t *testing.T) {
 	clientIP, proxyIP := net.IP{1, 2, 3, 4}, net.IP{127, 0, 0, 1}
-	msg := createTestMessage()
 
-	doRequest := func(t *testing.T, proxyAddr string) {
+	doRequest := func(t *testing.T, proxyAddr string, expectedClientIP net.IP) {
+		// Prepare the proxy server.
+		tlsConf, caPem := createServerTLSConfig(t)
+		dnsProxy := createTestProxy(t, tlsConf)
+
+		var gotAddr net.Addr
+		dnsProxy.RequestHandler = func(_ *Proxy, d *DNSContext) (err error) {
+			gotAddr = d.Addr
+
+			return dnsProxy.Resolve(d)
+		}
+
+		client := createTestHTTPClient(dnsProxy, caPem, false)
+
+		msg := createTestMessage()
+
 		dnsProxy.TrustedProxies = []string{proxyAddr}
 
 		// Start listening.
@@ -91,20 +92,17 @@ func TestHttpsProxyTrustedProxies(t *testing.T) {
 
 		resp := sendTestDoHMessage(t, client, msg, hdrs)
 		requireResponse(t, msg, resp)
+
+		ip, _ := netutil.IPAndPortFromAddr(gotAddr)
+		require.True(t, ip.Equal(expectedClientIP))
 	}
 
 	t.Run("success", func(t *testing.T) {
-		doRequest(t, proxyIP.String())
-
-		ip, _ := netutil.IPAndPortFromAddr(gotAddr)
-		assert.True(t, ip.Equal(clientIP))
+		doRequest(t, proxyIP.String(), clientIP)
 	})
 
 	t.Run("not_in_trusted", func(t *testing.T) {
-		doRequest(t, "127.0.0.2")
-
-		ip, _ := netutil.IPAndPortFromAddr(gotAddr)
-		assert.True(t, ip.Equal(proxyIP))
+		doRequest(t, "127.0.0.2", proxyIP)
 	})
 }
 
@@ -344,10 +342,17 @@ func sendTestDoHMessage(
 		req.Header.Set(k, v)
 	}
 
-	httpResp, err := client.Do(req)
+	httpResp, err := client.Do(req) // nolint:bodyclose
 	require.NoError(t, err)
 	testutil.CleanupAndRequireSuccess(t, httpResp.Body.Close)
 
+	require.True(
+		t,
+		httpResp.ProtoAtLeast(2, 0),
+		"the proto is too old: %s",
+		httpResp.Proto,
+	)
+
 	body, err := io.ReadAll(httpResp.Body)
 	require.NoError(t, err)
 
@@ -372,6 +377,8 @@ func createTestHTTPClient(dnsProxy *Proxy, caPem []byte, http3Enabled bool) (cli
 	var transport http.RoundTripper
 
 	if http3Enabled {
+		tlsClientConfig.NextProtos = []string{"h3"}
+
 		transport = &http3.RoundTripper{
 			Dial: func(
 				ctx context.Context,
@@ -395,10 +402,12 @@ func createTestHTTPClient(dnsProxy *Proxy, caPem []byte, http3Enabled bool) (cli
 			return dialer.DialContext(ctx, network, dnsProxy.Addr(ProtoHTTPS).String())
 		}
 
+		tlsClientConfig.NextProtos = []string{"h2", "http/1.1"}
 		transport = &http.Transport{
 			TLSClientConfig:    tlsClientConfig,
 			DisableCompression: true,
 			DialContext:        dialContext,
+			ForceAttemptHTTP2:  true,
 		}
 	}
 
